##
# Renew Splunk's self-signed server SSL certificates with a new expiration date
# but old private key.
# v1.0.1
#
# This script requires PowerShell 2.0 or later.
##

param(
  [switch]$help        = $false,
  [switch]$defaultCA   = $false,
  [switch]$liveCA      = $false,
  [string]$opensslConf = $env:OPENSSL_CONF,
  [switch]$serverCert  = $false,
  [string]$splunkHome  = $env:SPLUNK_HOME
)

########## Default settings ###################################################

$global:DAYS = 3650
$global:PASSWORD = "password"

$global:CA_COUNTRY = "US"
$global:CA_STATE = "CA"
$global:CA_CITY = "San Francisco"
$global:CA_ORGANIZATION = "Splunk"
$global:CA_COMMON_NAME = "SplunkCommonCA"
$global:CA_E_MAIL = "support@splunk.com"

$global:COMMON_NAME = "SplunkServerDefaultCert"

$global:SPLUNK_CERT_EXPIRATION_YEAR = 2016

########## Functions ##########################################################

$global:ME = "s-renewcerts.ps1"

function error( $msg, $status ) {
  $host.ui.WriteErrorLine( "$ME`: error: $msg" )
  Exit $status
}

function info( $msg ) {
  Write-Host "$ME`: $msg"
}

function assert_exists( $path ) {
  if ( -not (Test-Path -LiteralPath "$path") ) {
    error "$path not found" 11
  }
}

function extract_x509_first_cert( $from, $to ) {
  cat -LiteralPath "$from" -Delimiter "-----END CERTIFICATE-----" `
    -TotalCount 1 > "$to"
}

function extract_x509_private_key( $from, $to ) {
  assert_exists "$from"
  # We don't use "openssl pkey" since that's not available in older openssl.
  $begin = (Select-String -path "$from" -pattern "BEGIN.*PRIVATE" -List).LineNumber
  $end   = (Select-String -path "$from" -pattern "END.*PRIVATE" -List).LineNumber
  if ( -not ($begin -and $end) ) {
    error "$from doesn't contain a private key" 12
  }
  --$begin
  --$end
  Get-Content -path "$from" | Select-Object -Index ($begin..$end) | `
    Set-Content -LiteralPath "$to" -force
}

function get_x509_organization( $pem ) {
  assert_exists "$pem"
  $issuer = openssl x509 -in "$pem" -text -noout | `
    Select-String -SimpleMatch "Issuer"
  (("$issuer" -split ',')[3] -split '=')[1]
}

function get_x509_not_after_year( $pem ) {
  assert_exists "$pem"
  $not_after = openssl x509 -in "$pem" -text -noout | `
    Select-String -SimpleMatch "Not After"
  ("$not_after" -split '  *')[7]
}

function assert_is_Splunk_cert( $pem ) {
  $org = get_x509_organization "$pem"
  if ( "$org" -ne "$CA_ORGANIZATION" ) {
    error "$pem is not a Splunk-issued certificate" 13
  }
  get_x509_not_after_year $pem
}

function assert_Splunk_cert_is_new( $pem ) {
  $exp_year = assert_is_Splunk_cert "$pem"
  if ( $exp_year -le $SPLUNK_CERT_EXPIRATION_YEAR ) {
    error "$pem is not a renewed Splunk certificate" 14
  }
}

function assert_Splunk_cert_is_old( $pem ) {
  $exp_year = assert_is_Splunk_cert "$pem"
  if ( $exp_year -gt $SPLUNK_CERT_EXPIRATION_YEAR ) {
    error "$pem is already a renewed Splunk certificate" 15
  }
}

function backup( $file ) {
  $orig = "${file}.orig"
  # Make a backup only if there isn't one already.
  if ( (Test-Path -LiteralPath "$file") -and -not (Test-Path -LiteralPath "$orig") ) {
    cp -LiteralPath "$file" "$orig" -force
  }
}

function restore( $file ) {
  $orig = "${file}.orig"
  cp -LiteralPath "$orig" "$file" -force
}

function usage( $status ) {
  $usage = @"
usage: $ME [options]

Renew Splunk's self-signed SSL certificates.

optional arguments:
  -help                show this help message and exit
  -defaultCA           renew default CA certificates [default: false]
  -liveCA              renew live CA certificates [default: false]
  -opensslConf <path>  set/override value of %OPENSSL_CONF%
  -serverCert          renew server.pem certificate [default: false]
  -splunkHome <dir>    set/override value of %SPLUNK_HOME%

At least one of -defaultCA, -liveCA, or -serverCert is required;
order should be -defaultCA, -liveCA, -serverCert.
"@
  Write-Host "$usage"
  Exit $status
}

########## Process command-line ###############################################

if ( -not ($defaultCA -or $liveCA -or $serverCert) ) {
  usage 2
}

if ( -not ($opensslConf) ) {
  error "%OPENSSL_CONF% not set; set or use -opensslConf" 4
}

if ( -not (Test-Path -LiteralPath "$opensslConf") ) {
  error "%OPENSSL_CONF% ($opensslConf) does not exist" 5
}

if ( -not ($splunkHome) ) {
  error "%SPLUNK_HOME% not set; set or use -splunkHome" 6
}

if ( -not (Test-Path -LiteralPath "$splunkHome") ) {
  error "%SPLUNK_HOME% ($splunkHome) does not exist" 7
}

if ( -not (Test-Path -LiteralPath "$splunkHome" -PathType Container) ) {
  error "%SPLUNK_HOME% ($splunkHome) is not a directory" 8
}

########## Check for required commands ########################################

if ( -not (Get-Command "openssl" -ErrorAction SilentlyContinue) ) {
  error "openssl not found in %PATH% ($env:PATH)" 9
}

$env:OPENSSL_CONF = $opensslConf

########## Other variables ####################################################

$global:AUTH_DIR = "$splunkHome\etc\auth"
$global:CA_PEM = "$AUTH_DIR\ca.pem"
$global:CA_PEM_DEFAULT = "$CA_PEM.default"
$global:CACERT_PEM = "$AUTH_DIR\cacert.pem"
$global:CACERT_PEM_DEFAULT = "$CACERT_PEM.default"
$global:SERVER_PEM = "$AUTH_DIR\server.pem"

$global:CA_SUBJECT = "/C=$CA_COUNTRY/ST=$CA_STATE/L=$CA_CITY/O=$CA_ORGANIZATION/CN=$CA_COMMON_NAME/emailAddress=$CA_E_MAIL/"
$global:CERT_SUBJECT = "/CN=$COMMON_NAME/O=SplunkUser/"

$global:TMP_CERT_PEM = [System.IO.Path]::GetTempFileName()
$global:TMP_CSR_PEM  = [System.IO.Path]::GetTempFileName()
$global:TMP_KEY_PEM  = [System.IO.Path]::GetTempFileName()

trap {
  rm -LiteralPath "$TMP_CERT_PEM" -force
  rm -LiteralPath "$TMP_CSR_PEM"  -force
  rm -LiteralPath "$TMP_KEY_PEM"  -force
}

########## Renew default CA certificates ######################################

if ( $defaultCA ) {
  assert_Splunk_cert_is_old "$CA_PEM_DEFAULT"
  assert_Splunk_cert_is_old "$CACERT_PEM_DEFAULT"

  extract_x509_private_key "$CA_PEM_DEFAULT" "$TMP_KEY_PEM"

  ##
  # Create certificate signing request (CSR).
  ##
  openssl req `
    -new -key "$TMP_KEY_PEM" `
    -passin "pass:$PASSWORD" `
    -sha1 `
    -subj "$CA_SUBJECT" `
    -out "$TMP_CSR_PEM"

  ##
  # Self-sign the renewed certificate with the root CA and CSR.
  ##
  backup "$CACERT_PEM_DEFAULT"
  openssl x509 `
    -req -in "$TMP_CSR_PEM" `
    -passin "pass:$PASSWORD" `
    -days $DAYS `
    -sha1 `
    -extensions v3_ca `
    -signkey "$TMP_KEY_PEM" `
    -out "$CACERT_PEM_DEFAULT"

  ##
  # Recreate the new ca.pem.default.
  ##
  backup "$CA_PEM_DEFAULT"
  cat -LiteralPath "$CACERT_PEM_DEFAULT", "$TMP_KEY_PEM" -force | `
    Set-Content -LiteralPath "$CA_PEM_DEFAULT" -force

  info "renewed default CA certificates"
}

########## Renew live CA certificates #########################################

if ( $liveCA ) {
  assert_Splunk_cert_is_new "$CA_PEM_DEFAULT"
  assert_Splunk_cert_is_new "$CACERT_PEM_DEFAULT"

  assert_Splunk_cert_is_old "$CA_PEM"
  assert_Splunk_cert_is_old "$CACERT_PEM"

  backup "$CA_PEM"
  backup "$CACERT_PEM"
  Try {
    cp -LiteralPath "$CA_PEM_DEFAULT" "$CA_PEM" -force
    cp -LiteralPath "$CACERT_PEM_DEFAULT" "$CACERT_PEM" -force
  }
  Catch {
    restore "$CA_PEM"
    restore "$CACERT_PEM"
    error "renewing live certificates failed" 31
  }

  info "renewed live CA certificates"
}

########## Renew server.pem ###################################################

if ( $serverCert ) {
  assert_Splunk_cert_is_new "$CACERT_PEM"
  $exp_year = assert_is_Splunk_cert "$SERVER_PEM"

  extract_x509_first_cert "$SERVER_PEM" "$TMP_CERT_PEM"
  extract_x509_private_key "$SERVER_PEM" "$TMP_KEY_PEM"

  ##
  # Create a new server.pem with the original cert and private key, but the new
  # cacert.
  ##
  backup "$SERVER_PEM"
  Try {
    cat -LiteralPath "$TMP_CERT_PEM", "$TMP_KEY_PEM", "$CACERT_PEM" -force | `
      Set-Content -LiteralPath "$SERVER_PEM" -force
  }
  Catch {
    restore "$SERVER_PEM"
    error "creation of $SERVER_PEM failed" 41
  }

  info "renewed server certificate"
}

###############################################################################

if ( $liveCA -or $serverCert ) {
  info "splunk must be restarted for changes to take effect"
}

###############################################################################
# vim:set et sw=2 ts=2:
